# Copyright 2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyjoulescope_driver import Driver


def parser_config(p):
    """Joulescope info."""
    p.add_argument('--verbose', '-v',
                   action='store_true',
                   help='Display verbose information.')
    p.add_argument('--open', '-o',
                   choices=['defaults', 'restore'],
                   default='defaults',
                   help='The device open mode.  Defaults to "defaults".')
    p.add_argument('device_path',
                   nargs='*',
                   help='The target device for this command.')
    return on_cmd


def version_to_str(version):
    if isinstance(version, str):
        return version
    v_patch = version & 0xffff
    v_minor = (version >> 16) & 0xff
    v_major = (version >> 24) & 0xff
    return f'{v_major}.{v_minor}.{v_patch}'


class Info:

    def __init__(self, device_paths):
        self._device_paths = device_paths
        self._meta = {}
        self._values = {}

    def _on_pub(self, topic, value):
        self._values[topic] = value

    def _on_metadata(self, topic, value):
        if topic[-1] == '$':
            topic = topic[:-1]
        self._meta[topic] = value

    def run(self, args):
        with Driver() as d:
            d.log_level = args.jsdrv_log_level
            devices = d.device_paths()
            if not len(self._device_paths):
                self._device_paths = devices
            for device_path in self._device_paths:
                if device_path not in devices:
                    print(f'Device {device_path} not found in:')
                    s = "  \n".join(devices)
                    print(f'  {s}')
                    return 1
                self._meta.clear()
                self._values.clear()
                d.open(device_path, mode=args.open)
                fn = self._on_metadata  # use same bound method for unsubscribe
                d.subscribe(device_path, 'metadata_rsp_retain', fn)
                d.unsubscribe(device_path, fn)
                fn = self._on_pub  # use same bound method for unsubscribe
                d.subscribe(device_path, 'pub_retain', fn)
                d.unsubscribe(device_path, fn)
                if args.verbose:
                    print(f'{device_path} metadata:')
                    for key, value in self._meta.items():
                        subtopic = key[len(device_path) + 1:]
                        print(f'  {subtopic} {value}')
                    print(f'\n{device_path} values:')
                for key, value in self._values.items():
                    meta = self._meta.get(key, None)
                    if meta is not None:
                        fmt = meta.get('format', None)
                        if fmt == 'version':
                            value = version_to_str(value)
                    if args.verbose:
                        subtopic = key[len(device_path) + 1:]
                        print(f'  {subtopic} = {value}')
                    else:
                        print(f'{key} = {value}')
        return 0


def on_cmd(args):
    return Info(args.device_path).run(args)



# Copyright 2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyjoulescope_driver import Driver
import threading
import time


_quit = False


def _validate_timeout(v):
    v = float(v)
    if v == 0 or v >= 0.001:
        return v
    raise ValueError(f'Invalid timeout {v}')


def parser_config(p):
    """Threading demonstration for the pyjoulescope_driver.

    For example:
        python -m pyjoulescope_driver --log_level INFO --jsdrv_log_level info api_timeout --threads 10 --duration 10 --timeout 0.1
    """
    p.add_argument('--duration', '-d',
                   type=float,
                   default=1.0,
                   help='The duration in seconds to run the threading demonstration.')
    p.add_argument('--threads',
                   type=int,
                   default=10,
                   help='The number of threads to run simultaneously for API timeout testing.')
    p.add_argument('--timeout', '-t',
                   type=_validate_timeout,
                   default=0.25,
                   help='The API timeout in seconds.')
    return on_cmd


def _thread_run(state):
    global _quit
    index = state['index']
    d = state['driver']
    counter = 0
    while not _quit:
        try:
            d.publish('@/timeout', 0, timeout=state['timeout'])
            if state['timeout']:
                print('No timeout when expected')
                return 1
        except TimeoutError:
            pass  # expected
        counter += 1
    state['counter'] = counter
    print(f'{index} {counter}')


def on_cmd(args):
    global _quit
    with Driver() as d:
        d.log_level = args.jsdrv_log_level
        threads = []
        for idx in range(args.threads):
            state = {
                'name': f'jsdrv {idx}',
                'index': idx,
                'driver': d,
                'timeout': args.timeout,
                'counter': 0,
            }
            state['thread'] = threading.Thread(name=state['name'], target=_thread_run, args=[state])
            state['thread'].start()
            threads.append(state)

        t_end = time.time() + args.duration
        while time.time() < t_end:
            time.sleep(0.010)

        _quit = True
        counter = 0
        for thread in threads:
            thread['thread'].join()
            counter += thread['counter']
        print(f'Total calls: {counter}')

    return 0



# Copyright 2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import pyjoulescope_driver.release
import os
import shutil
import tempfile


MY_PATH = os.path.dirname(os.path.abspath(__file__))
RELEASE_PATH = os.path.join(MY_PATH, 'release')


class TestRelease(unittest.TestCase):

    def setUp(self):
        self._path = tempfile.mkdtemp(prefix='joulescope_driver')
        self._release_path = pyjoulescope_driver.release.release_path
        self._url_save = pyjoulescope_driver.release.url_save
        pyjoulescope_driver.release.release_path = lambda: self._path
        pyjoulescope_driver.release.url_save = self.url_save

    def tearDown(self):
        pyjoulescope_driver.release.url_save = self._url_save
        pyjoulescope_driver.release.release_path = self._release_path
        shutil.rmtree(self._path, ignore_errors=True)

    def url_save(self, url, filename):
        src = os.path.join(RELEASE_PATH, url)
        shutil.copy(src, filename)

    def test_program_path(self):
        path = self._release_path()
        self.assertTrue(os.path.isdir(path))
        self.assertEqual(self._path, pyjoulescope_driver.release.release_path())

    def test_release_get(self):
        r = pyjoulescope_driver.release.release_get('stable')
        self.assertTrue(r.startswith(b'joulescope_img_section'))



# Copyright 2018-2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
The Python time standard uses POSIX (UNIX) time which is defined relative
to Jan 1, 1970.  Python uses floating point seconds to express time.
This module defines a much simpler 64-bit fixed point integer for
representing time which is much friendlier for microcontrollers.
The value is 34Q30 with the upper 34 bits
to represent whole seconds and the lower 30 bits to represent fractional
seconds.  A value of 2**30 (1 << 30) represents 1 second.  This
representation gives a resolution of 2 ** -30 (approximately 1 nanosecond)
and a range of +/- 2 ** 33 (approximately 272 years).  The value is
signed to allow for simple arithmetic on the time either as a fixed value
or as deltas.

For more details, see jsdrv/time.h.
"""


import datetime
import numpy as np
import time as pytime


EPOCH = datetime.datetime(2018, 1, 1, tzinfo=datetime.timezone.utc).timestamp()  # seconds
SCALE = (1 << 30)
SECOND = SCALE
MINUTE = SECOND * 60
HOUR = MINUTE * 60
DAY = HOUR * 24
YEAR = DAY * 365  # common year
GREGORIAN_YEAR = int(365.2425 * DAY)   # https://en.wikipedia.org/wiki/Year

MILLISECOND = SCALE // 1_000
MICROSECOND = SCALE // 1_000_000
NANOSECOND = SCALE // 1_000_000_000  # bad precision!


def now():
    """Get the current timestamp.

    :return: The time64 representation of the current time.
    """
    return as_time64(pytime.time())


def as_time64(t):
    """Convert a time to a Joulescope timestamp.

    :param t: The time which can be:
        * datetime.datetime
        * python timestamp
        * integer Joulescope timestamp
    """
    if isinstance(t, int) or isinstance(t, np.int64):
        return t
    if isinstance(t, datetime.datetime):
        t = t.timestamp()
    else:
        t = float(t)
    return int((t - EPOCH) * SCALE)


def as_timestamp(t):
    """Convert a time to a python UTC timestamp.

    :param t: The time which can be:
        * datetime.datetime
        * python timestamp
        * integer Joulescope timestamp
    """
    t64 = as_time64(t)
    return (t64 / SCALE) + EPOCH


def as_datetime(t):
    """Convert a time to a python UTC timestamp.

    :param t: The time which can be:
        * datetime.datetime
        * python timestamp
        * integer Joulescope timestamp
    """
    t = as_timestamp(t)
    return datetime.datetime.fromtimestamp(t, tz=datetime.timezone.utc)


def filename(extension=None, t=None):
    """Construct a filename from the time.

    :param extension: The filename extension, such as '.png'.
        None (default) uses '.jls'.
    :param t: The time.  None (default) uses the current time.
    :return: The filename.
    """

    extension = '.jls' if extension is None else str(extension)
    if t is None:
        t = now()
    time_start = as_datetime(t)
    timestamp_str = time_start.strftime('%Y%m%d_%H%M%S')
    return f'{timestamp_str}{extension}'


def duration_to_seconds(d):
    """Convert a duration to float seconds.
    :param d: The duration specification, which is one of:
        * A string formatted as fz where f is a valid floating-point value
          and z is either omitted, 's', 'm', 'h', 'd'.
        * An integer in seconds.
        * A floating-point value in seconds.
    """
    if d is None:
        raise ValueError('cannot convert None')
    if isinstance(d, str):
        if not len(d):
            raise ValueError('cannot convert empty string')
        if d[-1] == 's':
            return float(d[:-1])
        elif d[-1] == 'm':
            return 60 * float(d[:-1])
        elif d[-1] == 'h':
            return 60 * 60 * float(d[:-1])
        elif d[-1] == 'd':
            return 60 * 60 * 24 * float(d[:-1])
        else:
            return float(d)
    else:
        return float(d)


def _local_offset():
    now = datetime.datetime.now()
    local_now = now.astimezone()
    local_tz = local_now.tzinfo
    print('hi')
    print(local_tz)


if __name__ == '__main__':
    _local_offset()



# Copyright 2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handle device programming for firmware and gateware.

For finer control on updating, use the jsdrv_util application.  For example:

jsdrv_util reset update1 && jsdrv_util mem_erase c/upd2 && jsdrv_util mem_write c/upd2 "C:/joulescope/js220/images/ctrl_release/updater2/js220_ctrl_updater2_1_0_1.img" && jsdrv_util reset update2
jsdrv_util reset update2 && jsdrv_util mem_erase c/upd1 && jsdrv_util mem_write c/upd1 "C:/joulescope/js220/images/ctrl_release/updater1/js220_ctrl_updater1_1_0_1.img" && jsdrv_util reset update1
jsdrv_util reset update1 && jsdrv_util mem_erase c/app && jsdrv_util mem_write c/app "C:/joulescope/js220/images/ctrl_release/app/js220_ctrl_app_1_0_1.img" && jsdrv_util reset app
jsdrv_util reset update1 && jsdrv_util mem_erase s/app1 && jsdrv_util mem_write s/app1 "C:/joulescope/js220/images/fpga_release/js220_fpga_1_0_1.img" && jsdrv_util reset app
"""


from pyjoulescope_driver import Driver
from pyjoulescope_driver.release import release_to_segments, \
    SUBTYPE_CTRL_APP, SUBTYPE_CTRL_UPDATER2, \
    SUBTYPE_CTRL_UPDATER1, SUBTYPE_SENSOR_FPGA, \
    TARGETS
import logging
import time


_log = logging.getLogger(__name__)
SUBTYPE_RESET = {
    SUBTYPE_CTRL_UPDATER1: 'update1',
    SUBTYPE_CTRL_UPDATER2: 'update2',
    SUBTYPE_CTRL_APP:      'app',
    SUBTYPE_SENSOR_FPGA:   None,
}


def version_to_str(version):
    if version is None:
        return '?.?.?'
    if isinstance(version, str):
        return version
    v_patch = version & 0xffff
    v_minor = (version >> 16) & 0xff
    v_major = (version >> 24) & 0xff
    return f'{v_major}.{v_minor}.{v_patch}'


class Programmer:

    def __init__(self, driver, device_path):
        self._driver = driver
        self._path = device_path
        if device_path[2] != '&':
            self._running_subtype = SUBTYPE_CTRL_APP
            self._path_app = device_path
            self._path_updater = device_path[:2] + '&' + device_path[2:]
        else:
            self._running_subtype = None  # either updater
            self._path_app = device_path[:2] + device_path[3:]
            self._path_updater = device_path

    @property
    def is_in_app(self):
        return self._path == self._path_app

    @property
    def version(self):
        return self._driver.query(self._path + '/c/fw/version')

    @property
    def fpga_version(self):
        return self._driver.query(self._path + '/s/fpga/version')

    @property
    def target_id(self):
        return self._driver.query(self._path + '/c/target')

    def _device_await(self, device_path, timeout=None):
        timeout = 5.0 if timeout is None else float(timeout)
        t_end = time.time() + timeout

        while True:
            device_paths = self._driver.device_paths()
            devices = [d for d in device_paths if d == self._path]
            if len(devices):
                if time.time() > t_end:
                    msg = f'timeout waiting for {self._path} removal'
                    _log.warning(msg)
                    raise TimeoutError(msg)
            else:
                break

        while True:
            device_paths = self._driver.device_paths()
            devices = [d for d in device_paths if d == device_path]
            if len(devices):
                self._path = devices[0]
                return
            if time.time() > t_end:
                msg = f'timeout waiting for {device_path} in {device_paths}'
                _log.warning(msg)
                raise TimeoutError(msg)
            time.sleep(0.1)

    def reset_to(self, subtype):
        if self._running_subtype == subtype:
            return
        reset_target = SUBTYPE_RESET[subtype]
        _log.info(f'reset to {reset_target} {subtype}')
        self._driver.publish(self._path + '/h/!reset', reset_target)
        self._driver.close(self._path)
        path = self._path_app if subtype == SUBTYPE_CTRL_APP else self._path_updater
        self._device_await(path)
        self._running_subtype = subtype
        self._driver.open(self._path)

    def segment_program(self, segment, progress, progress_msg, p1, p2):
        subtype = segment['subtype']
        target = TARGETS[subtype]
        _log.info('segment_program %s', target['name'])
        for attempt in range(10):
            attempt_msg = '' if attempt == 0 else f' (retry {attempt})'
            progress(p1, progress_msg + attempt_msg)
            if subtype == SUBTYPE_CTRL_UPDATER1:
                self.reset_to(SUBTYPE_CTRL_UPDATER2)
            elif subtype == SUBTYPE_CTRL_UPDATER2:
                self.reset_to(SUBTYPE_CTRL_UPDATER1)

            progress(p1, progress_msg + ' erase' + attempt_msg)
            self.publish(target['mem_region'] + '/!erase', 0, timeout=10)
            progress(p2, progress_msg + ' write' + attempt_msg)
            self.publish(target['mem_region'] + '/!write', segment['img'], timeout=10)
            reset_to = SUBTYPE_RESET[subtype]
            if reset_to is None:
                return None
            else:
                self.reset_to(subtype)
                version = self.version
                if self.target_id == target['target_id'] and version == segment['version']:
                    return version
        raise RuntimeError(f'Failed to program {target["name"]}')

    def query(self, topic):
        return self._driver.query(self._path + '/' + topic)

    def publish(self, topic, value, timeout=None):
        return self._driver.publish(self._path + '/' + topic, value, timeout=timeout)

    def open(self, mode=None, timeout=None):
        return self._driver.open(self._path, mode=mode, timeout=timeout)

    def close(self, timeout=None):
        return self._driver.close(self._path, timeout=timeout)


def release_program(driver: Driver, device_path: str, image: bytes,
                    force_program=None, progress=None):
    """Program a device with an official release.

    :param driver: The driver instance.
    :param device_path: The device path string for the target device.
    :param image: The binary release image to program.  See
        :func:`pyjoulescope_driver.release.release_get`.
    :param force_program: Force device programming for all segments.
        Normally, programming for segments with matching versions are skipped.
    :param progress: An optional callable(completion: float, msg: str) callback.
        When provided, this function will be called to provide progress updates.
        * completion: The completion status from 0.0 (not started) to 1.0 (done).
        * msg: A user-meaningful status message.
    """
    if progress is None:
        progress = lambda x, y: None
    segments = release_to_segments(image)
    progress(0.00, 'Check app version')
    programmer = Programmer(driver, device_path)
    if programmer.is_in_app:
        ctrl_app_version = programmer.version
    else:
        ctrl_app_version = None

    progress(0.02, 'Check updater2 version')
    try:
        programmer.reset_to(SUBTYPE_CTRL_UPDATER2)
        update2_version = programmer.version
    except TimeoutError:
        update2_version = None
        programmer.close()
        programmer.open()

    progress(0.05, 'Check updater1 version')
    try:
        programmer.reset_to(SUBTYPE_CTRL_UPDATER1)
        update1_version = programmer.version
    except TimeoutError:
        update1_version = None
        programmer.close()
        programmer.open()

    sensor_fpga_version = programmer.query('s/fpga/version')
    versions_before = [
        ['app', version_to_str(ctrl_app_version)],
        ['fpga', version_to_str(sensor_fpga_version)],
        ['update1', version_to_str(update1_version)],
        ['update2', version_to_str(update2_version)],
    ]
    v = ', '.join(['='.join(x) for x in versions_before])
    _log.info('Existing versions: %s', v)

    msg = 'Program updater2'
    progress(0.1, msg)
    segment = segments.get(SUBTYPE_CTRL_UPDATER2)
    if segment is not None and (force_program or segment['version'] != update2_version):
        update2_version = programmer.segment_program(segment, progress, msg, 0.1, 0.15)

    msg = 'Program updater1'
    progress(0.2, msg)
    segment = segments.get(SUBTYPE_CTRL_UPDATER1)
    if segment is not None and (force_program or segment['version'] != update1_version):
        update1_version = programmer.segment_program(segment, progress, msg, 0.2, 0.25)

    msg = 'Program sensor FPGA'
    progress(0.3, msg)
    segment = segments.get(SUBTYPE_SENSOR_FPGA)
    if segment is not None and (force_program or segment['version'] != sensor_fpga_version):
        programmer.segment_program(segment, progress, msg, 0.3, 0.55)

    msg = 'Program controller application'
    progress(0.65, msg)
    segment = segments.get(SUBTYPE_CTRL_APP)
    if segment is not None and (force_program or segment['version'] != ctrl_app_version):
        ctrl_app_version = programmer.segment_program(segment, progress, msg, 0.65, 0.9)
    else:
        programmer.reset_to(SUBTYPE_CTRL_APP)
    progress(0.97, 'Finalizing')
    sensor_fpga_version = programmer.query('s/fpga/version')
    programmer.close()
    
    versions_after = [
        ['app', version_to_str(ctrl_app_version)],
        ['fpga', version_to_str(sensor_fpga_version)],
        ['update1', version_to_str(update1_version)],
        ['update2', version_to_str(update2_version)],
    ]
    v = ', '.join(['='.join(x) for x in versions_after])
    _log.info('Updated to versions: %s', v)
    progress(1.0, 'Complete')
    return versions_before, versions_after



# Copyright 2022 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyjoulescope_driver import Driver
import time


def parser_config(p):
    """Display Joulescope measurement statistics."""
    p.add_argument('--verbose', '-v',
                   action='store_true',
                   help='Display verbose information.')
    p.add_argument('device_path',
                   nargs='*',
                   help='The target device for this command.')
    return on_cmd


def _on_statistics_value(topic, value):
    device = topic.split('/s/')[0]
    sample_id = value['time']['samples']['value'][0]
    i_avg, i_std, i_min, i_max = [value['signals']['current'][x]['value'] for x in ['avg', 'std', 'min', 'max']]
    v_avg, v_std, v_min, v_max = [value['signals']['voltage'][x]['value'] for x in ['avg', 'std', 'min', 'max']]
    p_avg, p_std, p_min, p_max = [value['signals']['power'][x]['value'] for x in ['avg', 'std', 'min', 'max']]
    charge = value['accumulators']['charge']['value']
    energy = value['accumulators']['energy']['value']
    print(f'{device},{sample_id},' +
          f'{i_avg},{i_std},{i_min},{i_max},' +
          f'{v_avg},{v_std},{v_min},{v_max},' +
          f'{p_avg},{p_std},{p_min},{p_max},' +
          f'{charge},{energy}')


def on_cmd(args):
    with Driver() as d:
        d.log_level = args.jsdrv_log_level
        devices = d.device_paths()
        if len(args.device_path):
            for device in args.device_path:
                if device not in devices:
                    print(f'device {device} not found')
                    return 1
            devices = args.device_path
        # display the column header
        print("#device,sampled_id," +
            "i_avg,i_std,i_min,i_max," +
            "v_avg,v_std,v_min,v_max," +
            "p_avg,p_std,p_min,p_max," +
            "charge,energy")
        for device in devices:
            d.open(device)
            if 'js110' in device:
                d.publish(device + '/s/i/range/select', 'auto')
                d.publish(device + '/s/i/ctrl', 'on')
            else:
                d.publish(device + '/s/i/range/mode', 'auto')
            d.publish(device + '/s/stats/ctrl', 1)
            d.subscribe(device + '/s/stats/value', 'pub', _on_statistics_value)
        try:
            while True:
                time.sleep(0.025)
        except KeyboardInterrupt:
            pass



# Copyright 2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Record streaming sample data to a JLS v2 file.
"""

import copy
import numpy as np
from pyjoulescope_driver import time64
import logging


_PYJLS_VERSION_MIN = (0, 6, 2)  # inclusive
_PYJLS_VERSION_MAX = (1, 0, 0)  # exclusive


try:
    from pyjls import Writer, SignalType, DataType, __version__
    _DTYPE_MAP = {
        'f32': DataType.F32,
        'u8': DataType.U8,
        'u4': DataType.U4,
        'u1': DataType.U1,
    }
except ImportError:
    Writer = None
    _DTYPE_MAP = {}


_SIGNALS = {
    'current': {
        'signal_type': 'f32',
        'units': 'A',
        'ctrl_topic': 's/i/ctrl',
        'data_topic': 's/i/!data',
    },
    'voltage': {
        'signal_type': 'f32',
        'units': 'V',
        'ctrl_topic': 's/v/ctrl',
        'data_topic': 's/v/!data',
    },
    'power': {
        'signal_type': 'f32',
        'units': 'W',
        'ctrl_topic': 's/p/ctrl',
        'data_topic': 's/p/!data',
    },
    'current_range': {
        'signal_type': 'u4',
        'units': '',
        'ctrl_topic': 's/i/range/ctrl',
        'data_topic': 's/i/range/!data',
    },
    'gpi[0]': {
        'signal_type': 'u1',
        'units': '',
        'ctrl_topic': 's/gpi/0/ctrl',
        'data_topic': 's/gpi/0/!data',
    },
    'gpi[1]': {
        'signal_type': 'u1',
        'units': '',
        'ctrl_topic': 's/gpi/1/ctrl',
        'data_topic': 's/gpi/1/!data',
    },
    'gpi[2]': {
        'signal_type': 'u1',
        'units': '',
        'ctrl_topic': 's/gpi/2/ctrl',
        'data_topic': 's/gpi/2/!data',
    },
    'gpi[3]': {
        'signal_type': 'u1',
        'units': '',
        'ctrl_topic': 's/gpi/3/ctrl',
        'data_topic': 's/gpi/3/!data',
    },
    'trigger_in': {
        'signal_type': 'u1',
        'units': '',
        'ctrl_topic': 's/gpi/7/ctrl',
        'data_topic': 's/gpi/7/!data',
    },
}


_SIGNAL_SHORT_MAP = [
    ('current', 'i'),
    ('voltage', 'v'),
    ('power', 'p'),
    ('current_range', 'r', 'current range'),
    ('gpi[0]', '0'),
    ('gpi[1]', '1'),
    ('gpi[2]', '2'),
    ('gpi[3]', '3'),
    ('trigger_in', 'T', 't'),
]


def _signal_name_map():
    m = {}
    for z in _SIGNAL_SHORT_MAP:
        signal_name = z[0]
        m[signal_name] = signal_name
        for n in z[1:]:
            m[n] = signal_name
    return m


class Record:
    """Record streaming sample data to a JLS v2 file.

    :param driver: The active driver instance.
    :param device_path: The device prefix path.
    :param signals: The list of signals to record.  None=['current', 'voltage']

    Call :meth:`open` to start recording and :meth:`close` to stop.
    """

    def __init__(self, driver, device_path, signals=None):
        if Writer is None:
            raise RuntimeError('pyjls package not found.  Install using:\n' +
                               '  pip3 install -U pyjls')
        pyjls_version = tuple([int(x) for x in __version__.split('.')])
        if pyjls_version < _PYJLS_VERSION_MIN or pyjls_version >= _PYJLS_VERSION_MAX:
            raise ImportError(f'Unsupported pyjls version {__version__}\n' +
                              f'  Require {_PYJLS_VERSION_MIN} <= pyjls version < {_PYJLS_VERSION_MAX}\n' +
                              '  pip3 install -U pyjls')
        self._utc_interval = time64.MINUTE
        self._log = logging.getLogger(__name__)
        self._wr = None
        self._data_map = {}
        self._driver = driver
        self._device_path = device_path
        self._on_data_fn = self._on_data  # bind and save for unsubscribe
        if signals is None:
            signals = ['current', 'voltage']
        elif isinstance(signals, str):
            signals = [s.strip() for s in signals.split(',')]
        m = _signal_name_map()
        signals = [m[s] for s in signals]

        signal_id = 0
        self._signals = {}
        for signal_name in signals:
            signal_id += 1
            signal = copy.deepcopy(_SIGNALS[signal_name])
            signal['name'] = signal_name
            signal['signal_id'] = signal_id
            signal['signal_type'] = _DTYPE_MAP[signal['signal_type']]
            signal['data_topic_abs'] = f"{self._device_path}/{signal['data_topic']}"
            signal['utc_next'] = None
            signal['utc'] = None
            self._signals[signal_name] = signal

    def open(self, filename):
        """Start the recording.

        :param filename: The filename for the recording.  Use
            time64.filename to produce a filename from timestamp.
        :return: self.
        """
        if self._wr is not None:
            self.close()
        self._data_map.clear()
        device_path = self._device_path
        self._wr = Writer(filename)
        _, model, serial_number = device_path.split('/')
        model = model.upper()
        self._wr.source_def(
            source_id=1,
            name=f'{model}-{serial_number}',
            vendor='Jetperch',
            model=model,
            version='',
            serial_number=serial_number,
        )

        for signal in self._signals.values():
            data_topic = signal['data_topic_abs']
            self._data_map[data_topic] = signal
            self._driver.subscribe(data_topic, ['pub'], self._on_data_fn)

        for signal in self._signals.values():
            ctrl_topic = signal['ctrl_topic']
            self._publish(ctrl_topic, 1, timeout=0)

        return self

    def _publish(self, topic, value, timeout=None):
        self._driver.publish(f'{self._device_path}/{topic}', value, timeout=timeout)

    def close(self):
        try:
            for signal in self._signals.values():
                if signal['utc'] is not None:
                    self._wr.utc(signal['signal_id'], *signal['utc'])
                ctrl_topic = signal['ctrl_topic']
                try:
                    self._publish(ctrl_topic, 0, timeout=0.25)
                except TimeoutError:
                    self._log.warning('Timed out in publish: %s <= 0', ctrl_topic)
                except Exception:
                    self._log.exception('Exception in publish: %s <= 0', ctrl_topic)
            for signal in self._signals.values():
                self._driver.unsubscribe(signal['data_topic_abs'], self._on_data_fn)
        finally:
            self._wr.close()
            self._wr = None

    def _on_data(self, topic, value):
        signal = self._data_map[topic]
        decimate_factor = value['decimate_factor']
        signal_id = signal['signal_id']
        sample_id = value['sample_id']
        sample_id = sample_id // decimate_factor
        if signal['utc_next'] is None:
            signal['sample_id_next'] = 0
            self._wr.signal_def(
                signal_id=signal['signal_id'],
                source_id=1,
                signal_type=SignalType.FSR,
                data_type=signal['signal_type'],
                sample_rate=value['sample_rate'] // decimate_factor,
                name=signal['name'],
                units=signal['units'],
            )
            self._wr.utc(signal_id, sample_id, value['utc'])
            signal['utc_next'] = value['utc'] + self._utc_interval

        if value['utc'] >= signal['utc_next']:
            self._wr.utc(signal_id, sample_id, value['utc'])
            signal['utc_next'] += self._utc_interval
            signal['utc'] = None
        elif sample_id:
            signal['utc'] = (sample_id, value['utc'])

        x = value['data']
        if len(x):
            x = np.ascontiguousarray(x)
            self._wr.fsr_f32(signal_id, sample_id, x)
        signal['sample_id_next'] += len(x)



# Copyright 2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyjoulescope_driver import Driver, Record, time64
import time


def parser_config(p):
    """Capture streaming samples to a JLS v2 file."""
    p.add_argument('--verbose', '-v',
                   action='store_true',
                   help='Display verbose information.')
    p.add_argument('--duration',
                   type=time64.duration_to_seconds,
                   help='The capture duration in float seconds. '
                        + 'Add a suffix for other units: s=seconds, m=minutes, h=hours, d=days')
    p.add_argument('--frequency', '-f',
                   type=int,
                   help='The sampling frequency in Hz.')
    p.add_argument('--open', '-o',
                   choices=['defaults', 'restore'],
                   default='defaults',
                   help='The device open mode.  Defaults to "defaults".')
    p.add_argument('--serial_number',
                   help='The serial number of the Joulescope for this capture.')
    p.add_argument('--set',
                   default=[],
                   action='append',
                   help='Set a parameter using "topic=value"')
    p.add_argument('--signals',
                   default='current,voltage',
                   help='The comma-separated list of signals to capture which include '
                        + 'current, voltage, power, current_range, gpi[0], gpi[1], gpi[2], gpi[3], trigger_in. '
                        + 'You can also use the short form i, v, p, r, 0, 1, 2, 3, T '
                        + 'Defaults to current,voltage.')
    p.add_argument('filename',
                   nargs='?',
                   default=time64.filename(),
                   help='The JLS filename to record. '
                        + 'If not provided, construct filename from current time.')
    return on_cmd


def on_cmd(args):
    with Driver() as d:
        d.log_level = args.jsdrv_log_level
        if args.serial_number is not None:
            device_paths = [p for p in d.device_paths() if p.lower().endswith(args.serial_number.lower())]
        else:
            device_paths = d.device_paths()
        if len(device_paths) == 0:
            print('Device not found')
            return
        elif len(device_paths) > 1:
            print(f'Selecting {device_paths[0]}')
        device_path = device_paths[0]

        if args.verbose:
            print(f'Open device: {device_path}')
        d.open(device_path, mode=args.open)
        if args.open == 'defaults':
            if 'js110' in device_path:
                pre_args = [
                    's/i/range/select=auto',
                    's/v/range/select=15 V'
                ]
            else:
                pre_args = [
                    's/i/range/mode=auto',
                    's/v/range/select=15 V',
                    's/v/range/mode=manual',  # as of 2023-03-24, auto not working well
                ]
        try:  # configure the device
            fs = args.frequency
            if fs is None:
                fs = 2_000_000 if 'js110' in device_path else 1_000_000
            else:
                fs = int(fs)
            d.publish(f'{device_path}/h/fs', fs)
        except Exception:
            print('failed to configure device')
            d.close(device_path)
            return 1

        for set_arg in pre_args + args.set:
            try:
                topic, value = set_arg.split('=')
                d.publish(f'{device_path}/{topic}', value)
            except Exception:
                print(f'Parameter set failed for {set_arg}')
                raise

        wr = Record(d, device_path, args.signals)
        if args.verbose:
            print(f'Record to file: {args.filename}')
        print('Start recording.  Press CTRL-C to stop.')
        wr.open(args.filename)
        t_stop = None if args.duration is None else time.time() + args.duration
        try:
            while t_stop is None or t_stop > time.time():
                time.sleep(0.010)
            if args.verbose:
                print(f'Record complete due to duration')
        except KeyboardInterrupt:
            if args.verbose:
                print(f'Record complete due to user CTRL-C')
        finally:
            wr.close()
            d.close(device_path)
    if args.verbose:
        print(f'Record complete')
    return 0



# Copyright 2022-2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyjoulescope_driver import Driver
import time


def parser_config(p):
    """Display Joulescope general-purpose input values."""
    p.add_argument('--verbose', '-v',
                   action='store_true',
                   help='Display verbose information.')
    return on_cmd

def _query_gpi_value(d, device):
    gpi_value = None

    def on_gpi_value(topic, value):
        nonlocal gpi_value
        gpi_value = value

    d.subscribe(f'{device}/s/gpi/+/!value', 'pub', on_gpi_value)
    d.publish(f'{device}/s/gpi/+/!req', 0)
    t_start = time.time()
    while gpi_value is None:
        if time.time() - t_start > 1.0:
            raise RuntimeError('_query_gpi_value timed out')
        time.sleep(0.001)
    d.unsubscribe(f'{device}/s/gpi/+/!value', on_gpi_value)
    return gpi_value

def on_cmd(args):
    with Driver() as d:
        d.log_level = args.jsdrv_log_level
        for device in d.device_paths():
            try:
                d.open(device, 'restore')
                gpi = _query_gpi_value(d, device)
                print(f'{device}: 0x{gpi:02x}')
                d.close(device)
            except Exception:
                print(f'{device} unavailable')



#!/usr/bin/env python3
# Copyright 2018-2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Joulescope driver python setuptools module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""


# Always prefer setuptools over distutils
import setuptools
from setuptools import Command
from setuptools.errors import ExecError
from setuptools.command.build_py import build_py
import numpy as np
import os
import platform
os.environ['pyjoulescope_driver_setup'] = '1'
from pyjoulescope_driver.release import releases_get_from_network


MYPATH = os.path.dirname(os.path.abspath(__file__))
VERSION_PATH = os.path.join(MYPATH, 'pyjoulescope_driver', 'version.py')
C_INCS = [
    os.path.join(MYPATH, 'include'),
    os.path.join(MYPATH, 'include_private'),
    os.path.join(MYPATH, 'third-party', 'tinyprintf'),
    np.get_include(),
]

try:
    from Cython.Build import cythonize
    USE_CYTHON = os.path.isfile(os.path.join(MYPATH, 'pyjoulescope_driver', 'binding.pyx'))
except ImportError:
    USE_CYTHON = False


about = {}
with open(VERSION_PATH, 'r', encoding='utf-8') as f:
    exec(f.read(), about)

extra_compile_args = []
extra_link_args = []
posix_sources = [
    'src/backend/posix.c',
    'src/backend/libusb/backend.c',
    'src/backend/libusb/msg_queue.c',
    'third-party/libusb/libusb/core.c',
    'third-party/libusb/libusb/descriptor.c',
    'third-party/libusb/libusb/hotplug.c',
    'third-party/libusb/libusb/io.c',
    'third-party/libusb/libusb/strerror.c',
    'third-party/libusb/libusb/sync.c',
    'third-party/libusb/libusb/os/events_posix.c',
    'third-party/libusb/libusb/os/threads_posix.c',
]


if platform.system() == 'Windows':
    sources = [
        'src/backend/winusb/backend.c',
        'src/backend/winusb/device_change_notifier.c',
        'src/backend/winusb/msg_queue.c',
        'src/backend/windows.c',
    ]
    libraries = ['Setupapi', 'Winusb', 'user32']
    extra_compile_args = []
elif 'armv7' in platform.machine():
    sources = posix_sources
    libraries = ['pthread', 'm']
    extra_compile_args = []
elif platform.processor() == 'aarch64' or platform.machine() == 'arm64':
    sources = posix_sources
    libraries = ['pthread', 'm']
    extra_compile_args = []
else:
    sources = posix_sources
    libraries = ['pthread', 'm']
    extra_compile_args = []


if platform.system() == 'Darwin':
    sources += [
        'third-party/libusb/libusb/os/darwin_usb.c'
    ]
    C_INCS.extend([
        os.path.join(MYPATH, 'third-party/libusb/libusb'),
        os.path.join(MYPATH, 'third-party/libusb/include/macos'),
    ])
    libraries += ['objc']
    extra_link_args = [
                  '-framework', 'IOKit',
                  '-framework', 'CoreFoundation',
                  '-framework', 'Security']
elif platform.system() == 'Linux':
    sources += [
        'third-party/libusb/libusb/os/linux_udev.c',
        'third-party/libusb/libusb/os/linux_usbfs.c',
    ]
    C_INCS.extend([
        os.path.join(MYPATH, 'third-party/libusb/libusb'),
        os.path.join(MYPATH, 'third-party/libusb/include/linux'),
    ])
    libraries += ['udev']

ext = '.pyx' if USE_CYTHON else '.c'
extensions = [
    setuptools.Extension('pyjoulescope_driver.binding',
                         sources=[
                                     'pyjoulescope_driver/binding' + ext,
                                     'src/buffer.c',
                                     'src/buffer_signal.c',
                                     'src/cstr.c',
                                     'src/devices.c',
                                     'src/downsample.c',
                                     #'src/emu.c',
                                     #'src/emulated.c',
                                     'src/error_code.c',
                                     'src/js110_cal.c',
                                     'src/js110_sample_processor.c',
                                     'src/js110_stats.c',
                                     'src/js110_usb.c',
                                     'src/js220_i128.c',
                                     'src/js220_params.c',
                                     'src/js220_usb.c',
                                     'src/js220_stats.c',
                                     'src/jsdrv.c',
                                     'src/json.c',
                                     'src/log.c',
                                     'src/pubsub.c',
                                     'src/meta.c',
                                     'src/sample_buffer_f32.c',
                                     'src/statistics.c',
                                     'src/time.c',
                                     'src/topic.c',
                                     'src/union.c',
                                     'src/version.c',
                                     'third-party/tinyprintf/tinyprintf.c',
                                     ] + sources,
                         include_dirs=C_INCS,
                         libraries=libraries,
                         extra_compile_args=[] + extra_compile_args,
                         extra_link_args=extra_link_args,
                         ),
]

if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions, compiler_directives={'language_level': '3'})  # , annotate=True)


# Get the long description from the README file
with open(os.path.join(MYPATH, 'README.md'), 'r', encoding='utf-8') as f:
    long_description = f.read()


class BuildPyCommand(build_py):
    """Custom build command."""

    def run(self):
        releases_get_from_network(force_download=True, dist_save=True)
        build_py.run(self)


class CustomBuildDocs(Command):
    """Custom command to build docs locally."""

    description = 'Build docs.'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # sphinx-build -b html docs build\docs_html
        # defer import so not all setups require sphinx
        from sphinx.application import Sphinx
        from sphinx.util.console import nocolor, color_terminal
        nocolor()
        source_dir = os.path.join(MYPATH, 'doc', 'sphinx')
        target_dir = os.path.join(MYPATH, 'build', 'docs_html')
        doctree_dir = os.path.join(target_dir, '.doctree')
        app = Sphinx(source_dir, source_dir, target_dir, doctree_dir, 'html')
        app.build()
        if app.statuscode:
            raise ExecError(
                'caused by %s builder.' % app.builder.name)


setuptools.setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=about['__url__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    license=about['__license__'],

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',

        # Pick your license as you wish
        'License :: OSI Approved :: Apache Software License',

        # Operating systems
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 11',
        'Operating System :: MacOS',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        # Supported Python versions
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',

        # Topics
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: Software Development :: Testing',
        'Topic :: System :: Hardware :: Hardware Drivers',
        'Topic :: Utilities',
        'Topic :: System :: Hardware :: Universal Serial Bus (USB)',
    ],

    keywords='Joulescope',

    packages=setuptools.find_packages(exclude=['native', 'docs', 'test', 'dist', 'build']),
    ext_modules=extensions,
    cmdclass={
        'docs': CustomBuildDocs,
        'build_py': BuildPyCommand,
    },
    include_dirs=[],
    include_package_data=True,

    # See https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='~=3.8',

    # See https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'numpy>=1.23',
        "pywin32; platform_system=='Windows'",
        'requests',
    ],

    entry_points={
        'console_scripts': [
            'pyjoulescope_driver=pyjoulescope_driver.__main__:run',
        ],
    },

    project_urls={
        'Bug Reports': 'https://github.com/jetperch/jls/issues',
        'Funding': 'https://www.joulescope.com',
        'Twitter': 'https://twitter.com/joulescope',
        'Source': 'https://github.com/jetperch/jls/',
    },
)
